"""Provides classes to build simple data tables.

"""
# Standard Library Imports
import copy
import math
import sys

# Local Package Imports
from ..utils import simple_csv


class ColumnStatistics(object):
    """Calculates statistics for a table column.

    """
    def __init__(self, index_id, column_id):
        """Initializes the class.

        When initializing this class, the 'column_id' parameter that  
        specifies the table column is provided, along with the 
        'index_id' parameter, which matches the 'index_id; attribute 
        of a table.

        Parameters
        ----------
        index_id : str
            The ID value of the column used to index the table.
        column_id : str
            The ID value of the column that will be used to generate 
            statistics.

        """
        setattr(self, '__index_id', index_id)
        setattr(self, '__column_id', column_id)

        return None

    def __is_number(self, value):
        # This method deterimines if the given value is either a number 
        # or can be converted to a number.  Any string values, even 
        # those that can be parsed into a number will not be considered 
        # a number.
        is_number = False

        if type(value) in (int, long, float, complex):
            is_number = True

        elif type(value) != str:

            try:
                x = float(value)

            except:
                x = None

            if x == float('nan'):
                x = None

            if x is not None:
                is_number = True

        return is_number

    def get_column_id(self):
        """Return the ID value of the table column.

        Returns
        -------
        str
          The ID value of the column used to generate statistics.

        """
        return getattr(self, '__column_id', None)

    def get_count(self):
        """Return the total count of rows processed by this class.

        Returns
        -------
        int
            The total number of rows processed.

        """
        return getattr(self, '__count', 0)

    def get_index_id(self):
        """Return the ID of the columns used to index the table.

        The 'index_id' value specifies which column in the table is 
        used to index row objects.

        Returns
        -------
        str
            The ID value of the column used to index the table.

        """
        return getattr(self, '__index_id', None)

    def get_max(self):
        """Returns the max value of the table column.

        Note
        ----
        While the return type of this method is 'float', any object 
        that implements numerical comparisons may be returned.

        Returns
        -------
        float
            The max value in a table column.

        """
        return getattr(self, '__max', None)

    def get_max_row(self):
        """Returns the table row containing the max value.

        Returns
        -------
        dict
            The row that contains the max value.

        """
        return getattr(self, '__max_row', None)

    def get_mean(self):
        """Returns the mean value of the table column.

        Note
        ----
        While the return type of this method is 'float', any object 
        that implements numerical comparisons may be returned.

        Returns
        -------
        float
            The mean value of the table column.

        """
        return getattr(self, '__mean', None)

    def get_min(self):
        """Returns the min value of the table column.

        Note
        ----
        While the return type of this method is 'float', any object 
        that implements numerical comparisons may be returned.

        Returns
        -------
        float
            The min value of the table column.

        """
        return getattr(self, '__min', None)

    def get_min_row(self):
        """Returns the table row containing the min value.

        Returns
        -------
        dict
            The row that contains the min value.

        """
        return getattr(self, '__min_row', None)

    def get_nncount(self):
        """Returns the number of non-numeric values in the column.

        Returns
        -------
        int
            The number of non-numeric values in the table column.

        """
        return getattr(self, '__nncount', 0)

    def get_total(self):
        """Returns the sum of the values in the column.

        Returns
        -------
        float
            The sum of all the values in the table column.

        """
        return getattr(self, '__total', None)

    def update(self, row):
        """Updates the statistical values from the given row.

        The row object passed to this method is a row from a Table 
        object.  The row must be a dictionary.

        Parameters
        ----------
        row : dict
            A row from a Table object.

        Returns
        -------
        None

        """
        column_id = self.get_column_id()
        value = row.get(column_id, None)

        if self.__is_number(value) is True:

            prev_max = self.get_max()
            prev_min = self.get_min()

            if prev_max is None:

                setattr(self, '__max', value)
                setattr(self, '__max_row', row)

            elif value > prev_max:

                setattr(self, '__max', value)
                setattr(self, '__max_row', row)

            if prev_min is None:

                setattr(self, '__min', value)
                setattr(self, '__min_row', row)

            elif value < prev_min:

                setattr(self, '__min', value)
                setattr(self, '__min_row', row)

            total = self.get_total()

            if total is None:
                total = value

            else:
                total += value

            setattr(self, '__total', total)

            count = self.get_count() + 1
            setattr(self, '__count', count)

            mean = total / float(count)
            setattr(self, '__mean', mean)

        else:

            nncount = self.get_nncount() + 1
            setattr(self, '__nncount', nncount)

        return None

    def updates(self, rows):
        """Updates the statistical values from the given list of rows.

        The rows in the list are from a Table object and must be given 
        as dictionary objects.

        Parameters
        ----------
        rows : list
            The list of rows from a table object, where each row is a 
            dictionary object.

        Returns
        -------
        None

        """
        for row in rows:
            self.update(row)

        return None


class Table(object):
    """Create a table of data using dictionary objects as rows.

    This class makes it simple to build a table of data by specifying 
    columns using a given 'column_ids' list.  Every element of the 
    'column_ids' list is a string and these elements determine the keys 
    in the table row dictionaries.  The keys in these dictionaries 
    can be thought of as the columns of the table.  

    The rows, given as dictionary objects, can be indexed by specifying 
    one of the 'column_ids' value as an index using the 'index_id' 
    parameter.  The rows are then sorted by the column specified by the 
    'index_id' parameter.

    To create a table of pre-determined size (row-wise), 'start', 'end', 
    and 'delta' parameters are provided.  The column specified by the 
    'index_id' parameter will contain the pre-generated index values of 
    the rows.  Rows are created where the index value is generated by 
    creating a row with the value of the 'start' parameter stored in 
    the column index.  Then, the value of the 'delta' parameter is 
    added to the first row's index value to create the next row.  This 
    continues until a row is created where the index value equals the 
    value of the 'end' parameter.  Currently, int and datetime objects 
    work best as index values.  Hoever, this is not enforced so use 
    with caution.

    The 'metadata' parameter provides context to the table.  This 
    parameter is not currently used internally.

    To offer more protection to attributes in the table from external 
    tampering, the optional parameter 'protected' is set to true by 
    default.  This mean that all rows, value, initialization 
    parameters are copied when setting or retrieving from the table.
    If 'protected' is set to False, then table attributes can be 
    modified from outside the instance. 

    Note
    ----
    This class is not built for performance.  It is intended to be used 
    with small to medium sized data sets, and is written to be as 
    simple as possible while providing a good balance of functionality.

    Attributes
    ----------
    csclass : ColumnStatistics
        Defines the class to use to calculate statistics.  This is 
        useful when extending the ''Table'' class.  An extended 
        ''ColumnStatistics'' class can be set to this attribute.

    """
    # This attribute allows any class that extends the Table class to 
    # specify their own statistics type.
    csclass = ColumnStatistics

    def __init__(self, column_ids = [], index_id = '', start = None, 
        end = None, delta = None, name = '', metadata = {}, protected = True):
        """The initialization method for the table.  

        The table may be initialized without specifying the 
        'column_ids' parameter, however the table won't be worth using 
        if that is the case.  The 'index_id' parameter is used to 
        index the rows of the table and to sort the rows of the table.  
        It is not required and may be be omitted.  If so, then rows are 
        unordered and listed in the order they are added.  The 'start', 
        'end', and 'delta' parameters are given only when needing to 
        create a table of pre-determined row length.  The 'name' and 
        'metadata' parameters are provided to describe the table.  Also 
        not required.  The 'protected' parameter will tell the table to 
        protect the values and attributes of the table from external 
        changes.

        Parameters
        ----------
        column_ids : list
            A list of str objects.  Each element of the list defines a 
            column in the table.
        index_id : str, optional
            Denotes the column used to index the rows of data in the 
            table.
        start : object, optional
            Used with the 'index_id', 'end', and 'delta' parameters. 
            Denotes the very first value in the column specified by 
            'index_id'.
        end : object, optional
            Used with the 'index_id', 'start', and 'delta' parameters. 
            Denotes the very last value in the column specified by 
            'index_id'.
        delta : object, optional
            Used with the 'index_id', 'start', and 'end' parameters. 
            Denotes the difference between values in the column 
            specified by the 'index_id' parameter.
        name : str, optional
            Specifies the name for the table.
        metadata : dict, optional
            Specifies any other information about the table.
        protected : bool, optional
            Specifies if the attributes in the table should be 
            protected from external tampering.  Default is True.

        """
        setattr(self, '__protected', protected)

        if self.is_protected() is True:

            column_ids = copy.deepcopy(column_ids)
            index_id = copy.deepcopy(index_id)
            start = copy.deepcopy(start)
            end = copy.deepcopy(end)
            delta = copy.deepcopy(delta)
            name = copy.deepcopy(name)
            metadata = copy.deepcopy(metadata)

        setattr(self, '__column_ids', column_ids)
        setattr(self, '__start', start)
        setattr(self, '__end', end)
        setattr(self, '__delta', delta)
        setattr(self, '__name', name)
        setattr(self, '__metadata', metadata)
        setattr(self, '__rows', [])

        if index_id in column_ids:

            setattr(self, '__index_id', index_id)
            setattr(self, '__rows_dict', {})

        if (self.get_index_id() != '') and (start is not None) and \
            (end is not None) and (delta is not None):

            try:
                self.__init_rows(start, end, delta)

            except Exception as e:
                sys.stderr.write(str(e) + '\n')

        return None

    def __init_row(self, index):
        # Creates an empty row as a dictionary object.  The given 
        # index_value argument is the row's index value.  This method 
        # is called from the __init_rows method.  The row is created 
        # with keys that match the values in the column_ids table 
        # attribute.
        row = {}

        for column_id in getattr(self, '__column_ids', []):

            if self.get_index_id() == column_id:
                row[column_id] = index

            else:
                row[column_id] = None

        return row

    def __init_rows(self, start, end, delta):
        # If, during table initialization, the table is provided 
        # index_id, start, end, and delta attributes, then, empty rows 
        # are created using the column specified by the index_id 
        # attribute.  The start object value is stored in the first 
        # row under the column used for indexing.  Then the following 
        # rows are created by adding the delta to the previous rows' 
        # index value.  Row initialization ends when the value of a 
        # row's index exceeds the value of the given end object.  The 
        # table is then locked, making new row creation extremely 
        # difficult.
        while start <= end:

            row = self.__init_row(start)
            self.set_row(row)

            start += delta

        if len(getattr(self, '__rows', [])) > 0:
            setattr(self, '__locked', True)

        return None

    def __row_as_list(self, row):
        # Takes a given row as a dictionary object and returns it as a 
        # list object.  The values from the dictionary are ordered 
        # in the same order as the column_ids table attribute.
        row_as_list = []

        for column_id in getattr(self, '__column_ids', []):
            row_as_list.append(row[column_id])

        return row_as_list

    def compare2(self, other):
        """Returns a new Table that compares this table to another.

        This table must have an 'index_id' attribute that is not an 
        empty string.  The other table's 'index_id' attribute must 
        match this tables 'index_id' attribute.  Further, the other 
        table must have at least one column ID value that matches a 
        value in the 'column_ids' attribute in this table.  If those 
        requirements are not satisfied, then a NoneType object is 
        returned.

        Parameters
        ----------
        other : Table
            A Table object that will be compared to this table.

        Returns
        -------
        Table
            A new Table object containing comparison columns between 
            this table and the given table.

        """
        return Table.comparison_table(self, other)

    @classmethod
    def comparison_table(cls, table1, table2):
        """Creates a new Table that is a comparison between two tables.

        The 'index_id' attribute in both tables must not be an empty 
        string and must match.  Further, at least one column ID value 
        in table2 must match a value in the 'column_ids' attribute of 
        table1.  If none of these requirements are satisfied, then a 
        NoneType object is returned.

        Parameters
        ----------
        table1 : Table
            The first table in the comparison.
        table2 : Table
            The second table in the comparison.

        Returns
        -------
        Table
            A new Table object that compares the contents of table1 to 
            the contents of table2.

        """
        # Create the index_id attribute for the comparison table.
        index_id = ''
        id1 = table1.get_index_id()
        id2 = table2.get_index_id()

        if (id1 != '') and (id2 != ''):
            if table1.get_index_id() == table2.get_index_id():
                index_id = table1.get_index_id()

        # Generate the names of the tables.  These names will serve as 
        # suffixes for the columns in the table.
        name1 = table1.get_name()
        name2 = table2.get_name()

        if not name1 or not name2:

            name1 = ' (1)'
            name2 = ' (2)'

        else:

            name1 = ' (' + name1 + ')'
            name2 = ' (' + name2 + ')'

        # Generate the new list of column ID values.  The list is 
        # composed of columns from the first table and columns from the 
        # second table that match the columns in the first.
        column_ids = [index_id]

        for c1 in table1.get_column_ids_ni():
            for c2 in table2.get_column_ids_ni():
                if c1 == c2:
                    column_ids.extend([c1 + name1, 
                        c1 + name2, 
                        c1 + ' diff', 
                        c1 + ' abs_diff', 
                        c1 + ' percent_diff'])

        if (index_id != '') and (len(column_ids) > 1):

            new_rows = []
    
            for row1 in table1.get_rows_as_dicts():
    
                index = row1.get(index_id, None)
                new_row = {index_id: index}
                row2 = table2.get_row_as_dict(index)

                if row2 is not None:
    
                    for c1 in table1.get_column_ids_ni():

                        v1 = row1.get(c1, None)
                        v2 = row2.get(c1, None)
                
                        sub_row[c1 + name1] = v1
                        sub_row[c1 + name2] = v2
                
                        if v1 and v2:
                
                            mean = (v1 + v2) / 2
                            diff = v1 - v2
                            abs_diff = abs(v1 - v2)
                            percent_diff = diff / mean * 100.0
                
                            sub_row[c1 + ' mean'] = mean
                            sub_row[c1 + ' diff'] = diff
                            sub_row[c1 + ' abs diff'] = abs_diff
                            sub_row[c1 + ' percent diff'] = percent_diff

                        new_row.update(sub_row)

                new_rows.append(new_row)

            table = cls(column_ids, index_id)

            for row in new_rows:
                table.set_row(row)

        else:
            table = None

        return table

    def get_column_ids(self):
        """Returns the list of the table's column ID values.

        Returns
        -------
        list
            A list of the table's column ID values.

        """
        column_ids = getattr(self, '__column_ids', [])

        if self.is_protected() is True:
            column_ids = copy.deepcopy(column_ids)

        return column_ids

    def get_column_ids_ni(self):
        """Returns the ID values for the non-index table columns.

        The ID value of the column used to index the table is excepted 
        from the list of ID values returned.

        Returns
        -------
        list
            A list of the table's column ID values, excepting the 
            ID of the column used for indexing.

        """
        column_ids = []

        for column_id in self.get_column_ids():
            if column_id != self.get_index_id():
                column_ids.append(column_id)

        return column_ids

    def get_column_statistics(self):
        """Returns statistics for the columns in this table.

        The max, min, and mean values are determined for each column 
        in this table using the ColumnStatistics class.  Only float or 
        int value types are used in the statistics calculations.

        Returns
        -------
        dict
            A dictionary of ColumnStatistics objects, where the keys 
            of the dictionary are column_id values.

        """
        statistics = {}

        index_id = self.get_index_id()
        column_ids = self.get_column_ids_ni()
        rows = self.get_rows_as_dicts()

        for column_id in column_ids:
            statistics[column_id] = self.csclass(index_id, column_id)

        for row in rows:
            for column_id in column_ids:
                statistics[column_id].update(row)

        return statistics

    def get_delta(self):
        """Returns the value of the 'delta' attribute.

        The 'delta' attribute is used to initialize rows when the table 
        is initialized with an 'index_id', 'start', 'end', and 'delta' 
        parameters.  Rows are initialized where the 'index_id' values 
        of the rows are uniformly stepped by the value of the 'delta' 
        attribute.

        Returns
        -------
        object
            The value of the difference between the initialized rows'
            'index_id' values.

        """
        delta = getattr(self, '__delta', None)

        if (delta is not None) and (self.is_protected() is True):
            delta = copy.deepcopy(delta)

        return delta

    def get_end(self):
        """Returns the value of the 'end' attribute.

        The 'end' attribute is used to initialize rows when the table 
        is initialized with an 'index_id', 'start', 'end', and 'delta' 
        parameters.  Rows are initialized where the last row that is 
        initialized has an 'index_id' value matching the end attribute.

        Returns
        -------
        object
            The last row's 'index_id' value, if the table was 
            initialized with an 'end' attribute.

        """
        end = getattr(self, '__end', None)

        if (end is not None) and (self.is_protected() is True):
            end = copy.deepcopy(end)

        return end

    def get_index_id(self):
        """Returns the 'index_id' attribute of this table.

        The 'index_id' specifies the column used to sort the rows of 
        this table.  Further, the values in the column specified by 
        the 'index_id' attribute must be unique.

        Returns
        -------
        str
            The 'index_id' attribute of this table.

        """
        return getattr(self, '__index_id', '')

    def get_metadata(self):
        """Returns the metadata for this table.

        The metadata should be contained in a dictionary.  If no 
        metadata was given during table initialization, then a NoneType 
        object will be returned.

        Returns
        -------
        dict
            The metadata of this table object, default None.

        """
        metadata = getattr(self, '__metadata', None)

        if (metadata is not None) and (self.is_protected() is True):
            metadata = copy.deepcopy(metadata)

        return metadata

    def get_name(self):
        """Returns the name of the table. 

        An empty string value is returned if name was not set when the 
        table was initialized.

        Returns
        -------
        str
            The table name.

        """
        return getattr(self, '__name', '')

    def get_row(self, index):
        """Returns a row in the table using the given index.

        The row is returned as a list object.  The elements of the 
        returned list are in the order of the 'column_ids' attribute of 
        this table.  The given index parameter may be either the index 
        into the list of rows in this table (eg. 15 would return the 
        fifteenth row) or it can be the value in the column used as the 
        table index.  For example, if the index parameter is 
        'datetime.date(2014, 8, 1)', then the row returned would have a 
        value of 'datetime.date(2014, 8, 1)' in the column used to index 
        the row.  

        Parameters
        ----------
        index : int or object
            The index value of the row to be retrieved.

        Returns
        -------
        list
            A list of values from a row in the table.

        """
        row = self.get_row_as_dict(index)

        if row is not None:
            row = self.__row_as_list(row)

        return row

    def get_row_as_dict(self, index):
        """Returns a row in the table using the given index.

        The row is returned as a dictionary object.  The given index 
        parameter may be either the index into the list of rows in this  
        table (eg. 15 would return the fifteenth row) or it can be the 
        value in the column used as the table index.  For example, if 
        the index parameter is 'datetime.date(2014, 8, 1)', then the 
        row returned would have a value of 'datetime.date(2014, 8, 1)' 
        in the column used to index the row.  

        Parameters
        ----------
        index : int or object
            The index value of the row to be returned.

        Returns
        -------
        dict
            A row from the table as a dictionary.

        """
        row = None

        if self.get_index_id() == '':

            if type(index) == int:

                rows = getattr(self, '__rows', [])

                if (index >= 0) and (index < len(rows)):
                    row = rows[index]

        else:

            rows_dict = getattr(self, '__rows_dict', {})
            row = rows_dict.get(str(index), None)

        if (row is not None) and (self.is_protected() is True):
            row = copy.deepcopy(row)

        return row

    def get_rows(self):
        """Returns the rows of this table in a list.

        The rows are returned in a list object, where each element of 
        the list is a row, and the row is also a list object.  The 
        values in the row list are in the order of the 'column_ids' 
        attribute of the table.  Further, if an 'index_id' attribute 
        was specified when the table was initialized, then the list of
        rows will be sorted by the values in the column specified by
        the 'index_id' attribute.

        Returns
        -------
        list
            A list of rows from this table, where each row in the 
            returned list is also a list.

        """
        _rows = getattr(self, '__rows', [])
        index = self.get_index_id()

        rows = []

        if index != '':
            _rows.sort(key = lambda x: x.get(index, None))

        for _row in _rows:
            rows.append(self.__row_as_list(_row))

        if self.is_protected() is True:
            rows = copy.deepcopy(rows)

        return rows

    def get_rows_as_dicts(self):
        """Returns the rows of this table in a list.

        The rows are returned in a list object, where each element of 
        the list is a row, and the row is a dictionary.  The dictionary 
        contains keys that match the values of the 'column_ids' 
        attribute of this table.  If an 'index_id' attribute was 
        specified when the table was initialized, then the list of rows 
        will be sorted by the key matching the 'index_id' attribute.

        Returns
        -------
        list
            A list of rows from this table, where each row in the 
            returned list is a dictionary.

        """
        rows = getattr(self, '__rows', [])
        index = self.get_index_id()

        if index != '':
            rows.sort(key = lambda x: x.get(index, None))

        if self.is_protected() is True:
            rows = copy.deepcopy(rows)

        return rows

    def get_rows_dict(self):
        """Returns a dictionary containing the table rows.

        If the 'index_id' attribute was not specified with this table  
        was initialized, then an empty dictionary will be returned.
        However, if the 'index_id' attribute was specified, then every 
        row in the table will be returned in a dictionary.  The keys 
        in the returned dictionary are the values of the column 
        specified by the 'index_id' attribute.

        Returns
        -------
        dict
            A dictionary containing all the rows in the table.

        """
        rows_dict = getattr(self, '__rows_dict', {})

        if self.is_protected() is True:
            rows_dict = copy.deepcopy(rows_dict)

        return rows_dict

    def get_start(self):
        """Returns the value of the start attribute.

        The start attribute is used to initialize rows when the table 
        is initialized with an 'index_id', 'start', 'end', and 'delta' 
        parameters.  Rows are initialized where the first row that is 
        initialized has an 'index_id' value matching the start 
        attribute.

        Returns
        -------
        object
            The first row's 'index_id' value, if the table is 
            initialized with a 'start' parameter.

        """
        start = getattr(self, '__start', None)

        if (start is not None) and (self.is_protected() is True):
            start = copy.deepcopy(start)

        return start

    def get_value(self, index, column_id):
        """Returns a value from a row in the table.

        This method finds the row with the given index value and 
        returns the value in the column specified by the 'column_id' 
        parameter.

        Parameters
        ----------
        index : object
            The index value of a row.
        column_id : str
            The ID of the column containing the value to be returned.

        Returns
        -------
        object
            The value from a row and column specified by the 'index' 
            and 'column_id' parameters.  Can be any type of Python 
            object.  Nonetype if a value is not found.

        """
        value = None

        if column_id in self.get_column_ids():
    
            if self.get_index_id() == '':

                if type(index) == int:
        
                    rows = getattr(self, '__rows', [])
        
                    if (index >= 0) and (index < len(rows)):
                        value = rows[index][column_id]
    
            else:
    
                rows_dict = getattr(self, '__rows_dict', {})
                _row = rows_dict.get(str(index), None)
    
                if _row is not None:
                    value = _row[column_id]

        if (value is not None) and (self.is_protected() is True):
            value = copy.deepcopy(value)

        return value

    def is_locked(self):
        """Returns whether this table will allow new rows.

        A table is locked when it is initialized with 'index_id', 
        'start', 'end', and 'delta' attributes.  With those parameters  
        provided, the table initializes the rows and then locks the 
        table so that no more rows can be added.  Row values can still 
        be modified, however.

        Returns
        -------
        bool
            True if the table is locked, false otherwise.

        """
        return getattr(self, '__locked', False)

    def is_protected(self):
        """Returns whether table values can be modified externally.

        This method determines whether values added or retrieved from 
        the table will be copied.  This is done to add an extra layer 
        of protection from the values having references external to the
        table, thus allowing external changes.

        Returns
        -------
        bool
            True if values are protected, false otherwise.

        """
        return getattr(self, '__protected', True)

    def set_row(self, row):
        """Adds or updates a row in this table.

        This method takes a dictionary object, where the keys in the 
        dictionary should match one or more values in the 'column_ids' 
        attribute of this table.  For the keys that match the 
        'column_ids' values, the values associated with those keys 
        are copied to a new dictionary and the dictionary is saved in 
        the table as a new row.  If an 'index_id' attribute was 
        specified during the table initialization, then this method 
        checks if the row exists.  If so, the current row is updated 
        with the new values from the given row.  If the table is 
        locked, then the values in the given row will not be saved if 
        the index value of the given row does not already exist in one 
        of the table rows.

        Parameters
        ----------
        row : dict
            A dictionary where the keys are ID values of the table 
            columns.

        Returns
        -------
        None

        """
        new_row = {}

        if self.is_protected() is True:
            row = copy.deepcopy(row)

        for i in getattr(self, '__column_ids', []):
            new_row[i] = row.get(i, None)

        index_id = self.get_index_id()
        index = str(new_row.get(index_id, ''))

        rows = getattr(self, '__rows', [])
        rows_dict = getattr(self, '__rows_dict', {})
        prev_row = rows_dict.get(index, None)

        if prev_row is not None:
            for key in prev_row.keys():
                prev_row[key] = new_row[key]

        else:

            if self.is_locked() is False:

                if index_id != '':
                    rows_dict[index] = new_row

                rows.append(new_row)

        return None

    def set_value(self, index, column_id, value):
        """Sets a value in a pre-existing row.

        This method will only work if the row with the given index 
        value exists.  Further, the value currently contained in the 
        column with the given 'column_id' parameter will be 
        overwritten.

        Parameters
        ----------
        index : int or object
            The index value of a pre-existing row.
        column_id : str
            The ID of the column where the new value is saved.

        Returns
        -------
        None

        """
        if column_id in getattr(self, '__column_ids', []):

            if self.is_protected() is True:
                value = copy.deepcopy(value)    

            if self.get_index_id() == '':

                if type(index) == int:
        
                    rows = getattr(self, '__rows', [])
        
                    if (index >= 0) and (index < len(rows)):
                        rows[index][column_id] = value
    
            else:
    
                rows_dict = getattr(self, '__rows_dict', {})
                _row = rows_dict.get(str(index), None)
    
                if _row is not None:
                    _row[column_id] = value

        return None

    def write_csv_row(self, count, row, prev_rows, **kwargs):
        """Writes a row of data to the CSV file.

        Any NoneType objects in the row are replaced with empty string.  
        Further, if an element of a row cannot be converted to a str 
        value, then an empty string is written in its place.

        Parameters
        ----------
        count : int
            The current row number being written.
        row : list
            The table row to be written to a CSV file.
        prev_rows : list
            An empty list. Provided for compatability with the 
            simple_csv module's callback function definition.

        Returns
        -------
        list
            A list of values that will be written as a row to a CSV 
            file.

        """
        new_row = []

        for i in row:

            try:

                if i is None:
                    new_row.append('')

                else:
                    new_row.append(str(i))

            except:
                new_row.append('')

        return new_row

    def write_to_csv(self, filename, callback = None):
        """Writes the table data to a CSV file.

        The table data is written to a CSV file specified by the given 
        filename parameter.  The values in the column_ids attribute are 
        written as the header in the file in the order that was given 
        when the table was initialized.  Then, each row is written to a 
        seperate line in the file.  The values in each row are written 
        to each line in the order of the column_ids attribute of the 
        table.

        Parameters
        ----------
        filename : str
            The name of the CSV file that will be written.
        callback : function, optional
            A function to prepare the row for writing, defaults to the 
            'write_csv_row' method in this class.

        Returns
        -------
        None

        """
        callback = callback

        if callback is None:
            callback = self.write_csv_row

        header = [self.get_column_ids()]
        rows = self.get_rows()

        simple_csv.write(filename, rows, callback = callback, header = header)

        return None


