"""Provides a datetime.tzinfo implementations.

Attributes
----------
mst : UTCOffset
    Represents the MST (Mountain Standard Time) timezone.
utc : UTCOffset
    Represents the UTC timezone.

"""
# Standard Library Imports
import datetime


class UTCOffset(datetime.tzinfo):
    """This class extends the datetime.tzinfo class. 

    The implementation of this class is kept simple, by not handling 
    daylight savings time.  The offset is specified by giving a 
    hourly value that is greater or equal to -12 and less then or 
    equal to 12.  The minutes are given as a value between 0 and 59, 
    inclusive.

    Attributes
    ----------
    hours : int
        The hours offset from UTC time.
    minutes : int
        The minutes offset from UTC time.

    """
    def __init__(self, hours, minutes):
        """Initialize the class.

        Parameters
        ----------
        hours : int
            The hours offset from UTC time.
        minutes: int
            The minutes offset from UTC time.
 
        """
        self.hours = hours
        self.minutes = minutes

        return None

    def __getinitargs__(self):
        # This method is required so that deep copy operations on 
        # instances of this class will work properly.
        return (self.hours, self.minutes)

    def __str__(self):
        # The magic method that prints the string value of the class 
        # instance.  The output is in this format: 'UTC +HH:MM' for 
        # positive offsets and 'UTC -HH:MM' for negative offsets.
        offset = 'UTC '

        if self.hours > 0:
            offset += '+'

        else:
            offset += '-'

        if self.hours > -10 and self.hours < 10 :
            offset += '0'

        offset += str(abs(self.hours)) + ':'

        if self.minutes < 10:
            offset += '0'

        offset += str(self.minutes)

        return offset

    def utcoffset(self, dt):
        """Returns the UTC offset of a datetime.

        Parameters
        ----------
        dt : datetime
            A datetime object that is used to determine the timedelta 
            between it and UTC time.

        Returns
        -------
        timedelta
            The delta between the current datetime and the UTC
            datetime.
        """
        return datetime.timedelta(hours = self.hours, minutes = self.minutes)

    def tzname(self, dt):
        """Return the string value of the datetime 'dt'.

        Parameters
        ----------
        dt : datetime
            A datetime object whose string value is returned.

        Returns
        -------
        str
            The string value of the given 'dt' parameter.

        """
        return str(dt)

    def dst(self, dt):
        """Returns a '0' timedelta.

        This class does not handle daylight savings time, so the 
        return value is a '0' delta.

        Parameters
        ----------
        dt : datetime
            The datetime to determine the daylight saving offset.

        Returns
        -------
        timedelta
            A timedelta object that represents a '0' delta.

        """
        return datetime.timedelta(0) 


# A UTC Timezone
utc = UTCOffset(0, 0)

# Mountain Standard Time
mst = UTCOffset(-7, 0)



