"""Provides functions to read and write CSV files.

Attributes
----------
QUOTE_ALL : int
    Defines an integer that is used to specify that a CSV file has 
    quotes around the values.
QUOTE_MINIMAL : int
    Defines an integer that is used to specify that a CSV file may have 
    quotes around the values.  For example, when values include the 
    delimiter.
QUOTE_NONE : int
    Defines an integer that is used to specify that a CSV file has 
    no quotes around any of the values.

"""
# Standard Library Imports
import copy
import csv
import sys
import urllib2

# Constants
QUOTE_ALL = csv.QUOTE_ALL
QUOTE_MINIMAL = csv.QUOTE_MINIMAL
QUOTE_NONE = csv.QUOTE_NONE


def _open(filename):

    try:

        if filename.startswith('http://'):
            f = urllib2.urlopen(filename)

        else:
            f = open(filename, 'rb')

    except Exception as e:

        sys.stderr.write(str(e) + '\n')
        f = None

    return f


def read(filename, delimiter = ',', quoting = QUOTE_NONE, callback = None, 
    use_prev_rows = 0, **kwargs):
    """Reads a CSV formatted file into a list of rows.

    This function can be used with only the 'filename' parameter being 
    passed.  However, extra functionality is provided to allow for 
    custom row parsing.  Further, any custom row parsing function 
    passed to this function can have access to ''n'' number of 
    previously parsed rows, if neccessary.

    Parameters
    ----------
    filename : str
        The name of the CSV formatted file.
    delimiter : str, optional
        The character used to seperate values, default: ','
    quoting : int, optional
        Denotes if the values are quoted, default: QUOTE_NONE.
    callback : callable, optional
        A custom row parsing function, default: None.
    use_prev_rows : int
        Specify the number of previous rows the callback function will 
        have access to, Default: 0.

    Returns
    -------
    list
        A list of parsed rows from a CSV formatted file.

    """
    func = lambda count, row, prev_rows, **y: row

    if callback is not None:
        if callable(callback) is True:
            func = callback

    f = _open(filename)

    if f is not None:

        reader = csv.reader(f, delimiter = delimiter, quoting = quoting)

        count = 0
        rows = []
        prev_rows = []

        for row in reader:

            try:
                row_from_func = func(count, row, prev_rows, **kwargs)

            except Exception as e:

                sys.stderr.write(str(e) + '\n')
                row_from_func = None

            if row_from_func is not None:

                rows.append(row_from_func)

                if use_prev_rows > 0:

                    prev_rows.append(copy.deepcopy(row_from_func))
                    prev_rows = prev_rows[-use_prev_rows:]

            count += 1

        f.close()

    else:
        rows = None

    return rows


def write(filename, rows, delimiter = ',', quoting = QUOTE_NONE, 
    callback = None, use_prev_rows = 0, header = None, footer = None, 
    **kwargs):
    """Write a CSV formatted file from a list of rows.

    Parameters
    ----------
    filename : str
        The name of the CSV formatted file.
    rows : list
        A list of rows that will be written as CSV formatted data.
    delimiter : str, optional
        The character to use to seperate the values in CSV file, 
        default: ','.
    quoting : int, optional
        Specifies if values should be quoted, default: QUOTE_NONE.
    callback : callable, optional
        A custom formatting function, default: None.
    use_prev_rows : int, optional
        Specify the number of previous rows the callback function will 
        have access to, default: 0.
    header : list, optional
        A list of values in lists to be written to the CSV file first, 
        default: None.
    footer : list, optional
        A list of values in lists to be written to the CSV file last, 
        default: None.

    Returns
    -------
    None 

    """
    func = lambda count, row, prev_rows, **y: row

    if callback is not None:
        if callable(callback) is True:
            func = callback

    f = open(filename, 'wb')
    writer = csv.writer(f, delimiter = delimiter, quoting = quoting)

    if header is not None:
        for header_row in header:
            writer.writerow(header_row)

    count = 0
    prev_rows = []

    for row in rows:

        row_from_func = func(count, row, prev_rows, **kwargs)

        if row_from_func is not None:

            writer.writerow(row_from_func)

            if use_prev_rows > 0:

                prev_rows.append(copy.deepcopy(row))
                prev_rows = prev_rows[-use_prev_rows:]

        count += 1

    if footer is not None:
        for footer_row in footer:
            writer.writerow(footer_row)

    f.close()

    return None


