"""Provides functions to download/upload files from/to a FTP server.

The FTP functionality provided in this module is not designed for 
secure FTP transactions.  Also, once the functions complete the file 
transfers, the FTP connection to the server is closed.  If more 
transactions are required, the functions create a new FTP object to 
use to reconnect.

"""
# Standard Library Imports
from ftplib import FTP
import os
import os.path


def download(host, port, user, passwd, remote_filenames = [], 
    local_dir = os.getcwd() + '/', remote_dir = '/', binary = True):
    """Downloads files via FTP from a remote server.

    Parameters
    ----------
    host : str
        Hostname of remote server.
    port : int
        The port on the remote server used by the FTP service.
    user : str
        The username to use to login to the remote server.
    passwd : str
        The user password used to login to the remote server.
    remote_filenames : list
        A list of filenames to download from the remote server.
    local_dir : str
        The local directory where the remote files will be saved.
    remote_dir : str
        The directory on the remote server where the files are located.
    binary : bool
        If True, download in binary mode, otherwise ascii mode is used.

    Returns
    -------
    None

    """
    ftp = FTP()
    gtg = True

    try:

        ftp.connect(host, port)
        ftp.login(user, passwd)
        ftp.cwd(remote_dir)

    except:
        gtg = False

    if gtg is True:

        actual_remote_filenames = ftp.nlst()

        retr = ftp.retrlines
        mode = 'w'
    
        if binary:
    
            retr = ftp.retrbinary
            mode = 'wb'
        
        if len(remote_filenames) == 0:
    
            for remote_filename in actual_remote_filenames:
    
                f = open(local_dir + remote_filename, mode)
                retr('RETR %s' % remote_filename, f.write)
                f.close()
    
        else:

            for remote_filename in remote_filenames:
    
                if remote_filename in actual_remote_filenames:

                    f = open(local_dir + remote_filename, mode)
                    retr('RETR %s' % remote_filename, f.write)
                    f.close()
    
    ftp.close()

    return None


def upload(host, port, user, passwd, local_filenames, remote_dir = '/',  
    binary = True):
    """Upload files via FTP to a remote server.

    Parameters
    ----------
    host : str
        Hostname of remote server.
    port : int
        The port on the remote server used by the FTP service.
    user : str
        The username to use to login to the remote server.
    passwd : str
        The user password used to login to the remote server.
    local_filenames : list
        A list of filenames to upload to the remote server.
    remote_dir : str
        The remote server directory where the uploaded files will be 
        stored.
    binary : bool
        If True, upload in binary mode, otherwise ascii mode is used.

    Returns
    -------
    None

    """
    ftp = FTP()
    gtg = True

    try:

        ftp.connect(host, port)
        ftp.login(user, passwd)
        ftp.cwd(remote_dir)

    except:
        gtg = False

    if gtg is True:

        stor = ftp.storlines
        mode = 'r'
    
        if binary:
    
            stor = ftp.storbinary
            mode = 'rb'
    
        for local_filename in local_filenames:
    
            if os.path.isfile(local_filename) is True:

                basename = os.path.basename(local_filename)
        
                f = open(local_filename, mode)
                stor('STOR %s' % basename, f)
                f.close()

    ftp.close()

    return None


